/*******************************************************************************
 * Copyright (c) 2003 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.contribution.xref.internal.ui.views;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.eclipse.contribution.xref.core.IDeferredXReference;
import org.eclipse.contribution.xref.core.IXReference;
import org.eclipse.contribution.xref.core.IXReferenceAdapter;
import org.eclipse.contribution.xref.internal.ui.XReferenceUIPlugin;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.ui.JavaElementLabelProvider;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.part.ViewPart;

/**
 * @author hawkinsh
 *  
 */
public class XReferenceView extends ViewPart implements ISelectionListener {

	public static final String ID =
		"org.eclipse.contribution.xref.internal.ui.views.XReferenceView";
	private static final String LINK_ID = ID + ".link";
	private static final String SELECTION_ID = ID + ".selection";

	private Action doubleClickAction;
	private Action toggleLinkingAction;
	private boolean linkingEnabled = true; // following selection?
	private IXReferenceAdapter lastSelection;
	private NavigationHistoryActionGroup navigationHistoryGroup;
	private TreeViewer viewer;
	private ViewContentProvider contentProvider;

	/*
	 * A class to help navigate the cross references
	 */
	public class TreeObject implements IAdaptable {
		private String name;
		private TreeParent parent;
		private Object data;

		public TreeObject(String name) {
			this.name = name;
		}
		public String getName() {
			return name;
		}
		public void setParent(TreeParent parent) {
			this.parent = parent;
		}
		public TreeParent getParent() {
			return parent;
		}
		public String toString() {
			return getName();
		}
		public Object getAdapter(Class key) {
			return null;
		}
		public Object getData() {
			return data;
		}
		public void setData(Object o) {
			data = o;
		}
	}

	/*
	 * A class to help navigate the cross references
	 */
	public class TreeParent extends TreeObject {
		private ArrayList children;
		public TreeParent(String name) {
			super(name);
			children = new ArrayList();
		}
		public void addChild(TreeObject child) {
			children.add(child);
			child.setParent(this);
		}
		public void removeChild(TreeObject child) {
			children.remove(child);
			child.setParent(null);
		}
		public TreeObject[] getChildren() {
			return (TreeObject[]) children.toArray(
				new TreeObject[children.size()]);
		}
		public boolean hasChildren() {
			return children.size() > 0;
		}
	}

	/*
	 * The content provider class is responsible for providing objects to the
	 * view. It can wrap existing objects in adapters or simply return objects
	 * as-is. These objects may be sensitive to the current input of the view,
	 * or ignore it and always show the same content (like Task List, for
	 * example).
	 */
	public class ViewContentProvider
		implements IStructuredContentProvider, ITreeContentProvider {

		private TreeParent invisibleRoot;
		private Object input = null;

		public void inputChanged(Viewer v, Object oldInput, Object newInput) {
			if (newInput != null) {
				input = newInput;
				initialize();
				if (v != null) { // guard to make testing easier
					v.refresh();
				}
			}
		}

		public void refresh() {
			// due to an underlying model change (evaluation of a deferred
			// cross reference) the content needs to be reevaluated.
			initialize();
		}

		public void dispose() {
			input = null;
			invisibleRoot = null;
		}

		public Object[] getElements(Object parent) {
			if (parent == null)
				return new Object[0];
			if (parent.equals(input)) {
				return getChildren(invisibleRoot);
			} else {
				return getChildren(parent);
			}
		}

		public Object getParent(Object child) {
			if (child instanceof TreeObject) {
				return ((TreeObject) child).getParent();
			}
			return null;
		}

		public Object[] getChildren(Object parent) {
			if (parent instanceof TreeParent) {
				return ((TreeParent) parent).getChildren();
			}
			return new Object[0];
		}

		public boolean hasChildren(Object parent) {
			if (parent instanceof TreeParent)
				return ((TreeParent) parent).hasChildren();
			return false;
		}

		private void initialize() {
			invisibleRoot = new TreeParent("");

			if ((input != null) && (input instanceof IXReferenceAdapter)) {
				IXReferenceAdapter xreferenceAdapter =
					(IXReferenceAdapter) input;
				TreeParent root = new TreeParent(xreferenceAdapter.toString());
				root.setData(xreferenceAdapter.getReferenceSource());
				invisibleRoot.addChild(root);

				Collection xrc = xreferenceAdapter.getXReferences();
				Iterator xri = xrc.iterator();
				while (xri.hasNext()) {
					IXReference xr = (IXReference) xri.next();
					TreeParent relName = new TreeParent(xr.getName());
					if (xr instanceof IDeferredXReference) {
						addEvaluateChild(relName, (IDeferredXReference) xr);
					}
					Iterator li = xr.getAssociates();
					while (li.hasNext()) {
						Object associate = li.next();
						TreeObject leaf = new TreeObject(associate.toString());
						leaf.setData(associate);
						relName.addChild(leaf);
					}
					root.addChild(relName);
				}
			}
		}

		private void addEvaluateChild(
			TreeParent parent,
			IDeferredXReference r) {
			TreeObject t =
				new TreeObject(
					XReferenceUIPlugin.getResourceString("EvaluateLabel"));
			t.setData(r);
			parent.addChild(t);
		}

	}

	public class ViewLabelProvider extends LabelProvider {

		// any element can be put into the data field of
		// a TreeObject. If it implements IWorkbenchAdapter,
		// it will be displayed correctly in the tree.

		private JavaElementLabelProvider labelProvider =
			new JavaElementLabelProvider(
				JavaElementLabelProvider.SHOW_POST_QUALIFIED
					| JavaElementLabelProvider.SHOW_OVERLAY_ICONS);

		public String getText(Object obj) {
			String ret = obj.toString();
			Object data = ((TreeObject) obj).getData();
			if ((data != null) && !(data instanceof IDeferredXReference)) {
				ret = labelProvider.getText(data);
			}
			return ret;
		}
		public Image getImage(Object obj) {
			Object data = ((TreeObject) obj).getData();
			if (data != null) {
				if (data instanceof IDeferredXReference) {
					return XReferenceUIPlugin.getDefault().getEvaluateImage();
				} else {
					return (labelProvider.getImage(data));
				}
			} else {
				return XReferenceUIPlugin.getDefault().getXReferenceImage();
			}
		}
	}

	class DoubleClickAction extends Action {

		public void run() {
			ISelection selection = viewer.getSelection();
			if (selection instanceof IStructuredSelection) {
				Object sel =
				((IStructuredSelection) selection).getFirstElement();
				Object data = ((TreeObject) sel).getData();
				if (data != null) {
					if (data instanceof IJavaElement) {
						revealInEditor((IJavaElement) data);
					} else if (data instanceof IDeferredXReference) {
						evaluateXReferences((IDeferredXReference) data);
					} else if (data instanceof IResource) {
						revealInEditor((IResource) data);
					}
				}
			}
		}

		private void revealInEditor(IJavaElement j) {
			try {
				IEditorPart p = JavaUI.openInEditor(j);
				JavaUI.revealInEditor(p, j);
				if (linkingEnabled) {
					navigationHistoryGroup.nowLeaving(viewer.getInput());
					viewer.setInput(j.getAdapter(IXReferenceAdapter.class));
				}
			} catch (Exception ex) {
				System.out.println(ex);
			}
		}

		private void revealInEditor(IResource r) {
			IMarker m;
			try {
				m = r.createMarker(IMarker.MARKER);
				IDE.openEditor(getSite().getPage(), m, true);
				//	getSite().getPage().openEditor(m);
			} catch (CoreException e) {
				e.printStackTrace();
			}
		}

		private void evaluateXReferences(IDeferredXReference xr) {
			try {
				Shell shell = getViewSite().getShell();
				new ProgressMonitorDialog(shell).run(true, true, xr);
				contentProvider.refresh();
				viewer.refresh();
				viewer.expandToLevel(3);
			} catch (InterruptedException intEx) {
				// user cancelled - this is ok...
			} catch (InvocationTargetException invEx) {
				System.err.println(
						"Something nasty here, "
						+ xr
						+ " could not be evaluated: "
						+ invEx);
			}
		}
	}
	
	/**
	 * The constructor
	 */
	public XReferenceView() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IWorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createPartControl(Composite parent) {
		viewer =
			new TreeViewer(parent, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL);
		navigationHistoryGroup = new NavigationHistoryActionGroup(viewer);
		contentProvider = new ViewContentProvider();
		viewer.setContentProvider(contentProvider);
		viewer.setLabelProvider(new ViewLabelProvider());
		//		viewer.setSorter(new NameSorter());
		viewer.setAutoExpandLevel(3);

		restorePersistedSettings();
		makeActions();
		hookDoubleClickAction();
		contributeToActionBars();
		
		IWorkbenchWindow window = getActiveWorkbenchWindow();

		if (window != null) {
			window.getSelectionService().addSelectionListener(this);
		}
	}

	/**
	 * Passing the focus request to the viewer's control.
	 */
	public void setFocus() {
		viewer.getControl().setFocus();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.ISelectionListener#selectionChanged(org.eclipse.ui.IWorkbenchPart,
	 *      org.eclipse.jface.viewers.ISelection)
	 */
	int testCounter = 0;
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		IAdaptable a = null;
		IXReferenceAdapter xra = null;
		if (selection instanceof IStructuredSelection) {
			IStructuredSelection structuredSelection =
				(IStructuredSelection) selection;
			Object first = structuredSelection.getFirstElement();
			if (first instanceof IAdaptable) {
				a = (IAdaptable) first;
			}
		}
		if (a != null) {
			xra = (IXReferenceAdapter) a.getAdapter(IXReferenceAdapter.class);
		}
		if (xra != null) {
			System.out.println(
				">> IXReferenceAdapter interface is supported: "
					+ testCounter++);
		} else {
			System.out.println(
				">> IXReferenceAdapter interface is NOT supported: "
					+ testCounter++);
		}
		if (xra != null) {
			lastSelection = xra;
			if (linkingEnabled) {
				viewer.setInput(xra);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.IWorkbenchPart#dispose()
	 */
	public void dispose() {
		super.dispose();
		IWorkbenchWindow window = getActiveWorkbenchWindow();

		if (window != null) {
			window.getSelectionService().removeSelectionListener(this);
		}
		persistSettings();
	}

	/*
	 * helper method to return the currently active workbench window (if there
	 * is one)
	 */
	private IWorkbenchWindow getActiveWorkbenchWindow() {
		return XReferenceUIPlugin
			.getDefault()
			.getWorkbench()
			.getActiveWorkbenchWindow();
	}

	protected boolean isLinkingEnabled() {
		return linkingEnabled;
	}

	protected void setLinkingEnabled(boolean isOn) {
		linkingEnabled = isOn;
		if (linkingEnabled && lastSelection != null) {
			viewer.setInput(lastSelection);
		}
	}

	private void persistSettings() {
		IPreferenceStore pstore =
			XReferenceUIPlugin.getDefault().getPreferenceStore();
		if (!pstore.contains(LINK_ID)) {
			pstore.setDefault(LINK_ID, true);
		}
		pstore.setValue(LINK_ID, linkingEnabled);
		IXReferenceAdapter xra = (IXReferenceAdapter) viewer.getInput();
		if (xra != null) {
			Object input = xra.getReferenceSource();
			String handle = null;
			if (input instanceof IResource) {
				handle = "R" + ((IResource) input).getFullPath().toString();
			} else if (input instanceof IJavaElement) {
				handle = "J" + ((IJavaElement) input).getHandleIdentifier();
			}
			if (handle != null) {
				pstore.setValue(SELECTION_ID, handle);
			}
		}
		XReferenceUIPlugin.getDefault().savePluginPreferences();
	}

	private void restorePersistedSettings() {
		IPreferenceStore pstore =
			XReferenceUIPlugin.getDefault().getPreferenceStore();
		if (pstore.contains(LINK_ID)) {
			linkingEnabled = pstore.getBoolean(LINK_ID);
		}
		if (pstore.contains(SELECTION_ID)) {
			String sel = pstore.getString(SELECTION_ID);
			String handle = sel.substring(1);
			IXReferenceAdapter xra = null;
			if (sel.startsWith("R")) {
				// its an IResource, handle is a path
				IPath p = new Path(handle);
				IResource r =
					XReferenceUIPlugin.getWorkspace().getRoot().findMember(
						handle);
				if (r != null) {
					xra =
						(IXReferenceAdapter) r.getAdapter(
							IXReferenceAdapter.class);
				}
			} else if (sel.startsWith("J")) {
				// its an IJavaElement
				IJavaElement j = JavaCore.create(handle);
				if (j != null) {
					xra =
						(IXReferenceAdapter) j.getAdapter(
							IXReferenceAdapter.class);
				}
			} else {
				// what the hell is it then?? - ignore
			}
			if (xra != null) {
				viewer.setInput(xra);
			}
		}
	}

	private void contributeToActionBars() {
		IActionBars bars = getViewSite().getActionBars();
		fillLocalPullDown(bars.getMenuManager());
		fillLocalToolBar(bars.getToolBarManager());
	}

	private void fillLocalPullDown(IMenuManager manager) {
		//manager.add(action2);
		manager.add(new Separator());
		manager.add(toggleLinkingAction);
	}
	
	private void fillLocalToolBar(IToolBarManager manager) {
		navigationHistoryGroup.addNavigationActions(manager);
		manager.add(new Separator());
		manager.add(toggleLinkingAction);
	}
	
	private void hookDoubleClickAction() {
		viewer.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent event) {
				doubleClickAction.run();
			}
		});
	}

	private void makeActions() {
		toggleLinkingAction = new ToggleLinkingAction(this);
//		collapseAllAction = new CollapseAllAction(this);
		doubleClickAction = new DoubleClickAction();
	}
	
}
