/* *******************************************************************
 * Copyright (c) 2002 Palo Alto Research Center, Incorporated (PARC).
 * All rights reserved. 
 * This program and the accompanying materials are made available 
 * under the terms of the Common Public License v1.0 
 * which accompanies this distribution and is available at 
 * http://www.eclipse.org/legal/cpl-v10.html 
 *  
 * Contributors: 
 *     PARC     initial implementation 
 * ******************************************************************/


package org.aspectj.ajdt.internal.compiler;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.aspectj.ajdt.internal.compiler.lookup.AjLookupEnvironment;
import org.aspectj.ajdt.internal.compiler.lookup.EclipseFactory;
import org.aspectj.ajdt.internal.compiler.lookup.InsideEclipseEclipseFactory;
import org.aspectj.ajdt.internal.compiler.problem.AjProblemReporter;
import org.aspectj.bridge.AbortException;
import org.aspectj.bridge.IMessage;
import org.aspectj.bridge.IMessageHandler;
import org.aspectj.bridge.MessageHandler;
import org.aspectj.bridge.IMessage.Kind;
import org.aspectj.weaver.World;
import org.aspectj.weaver.bcel.BcelWeaver;
import org.aspectj.weaver.bcel.BcelWorld;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.jdt.core.IJavaModelMarker;
import org.eclipse.jdt.core.compiler.IProblem;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.Compiler;
import org.eclipse.jdt.internal.compiler.DefaultErrorHandlingPolicies;
import org.eclipse.jdt.internal.compiler.ICompilerRequestor;
import org.eclipse.jdt.internal.compiler.IErrorHandlingPolicy;
import org.eclipse.jdt.internal.compiler.IProblemFactory;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.problem.DefaultProblem;
import org.eclipse.jdt.internal.compiler.problem.DefaultProblemFactory;
import org.eclipse.jdt.internal.compiler.problem.ProblemSeverities;
import org.eclipse.jdt.internal.core.builder.ClasspathLocation;
import org.eclipse.jdt.internal.core.builder.JavaBuilder;
import org.eclipse.jdt.internal.core.builder.NameEnvironment;
import org.eclipse.jdt.internal.core.builder.ProblemFactory;


public class AjCompiler extends Compiler {

	private EclipseFactory factory;
	private BcelWorld world;
	private BcelWeaver weaver;
	
	public AjCompiler(
		INameEnvironment environment,
		IErrorHandlingPolicy policy,
		Map settings,
		ICompilerRequestor requestor,
		IProblemFactory problemFactory,
		IMessageHandler handler,
		boolean inEclipse) {
		super(environment, policy, settings, requestor, problemFactory);
		init(environment, problemFactory, handler, inEclipse);
	}

	public AjCompiler(
		INameEnvironment environment,
		IErrorHandlingPolicy policy,
		Map settings,
		ICompilerRequestor requestor,
		IProblemFactory problemFactory,
		boolean parseLiteralExpressionsAsConstants,
		IMessageHandler handler,
		boolean inEclipse) {
		super(
			environment,
			policy,
			settings,
			requestor,
			problemFactory,
			parseLiteralExpressionsAsConstants);
		init(environment, problemFactory, handler, inEclipse);
	}
	
	/**
	 * In addition to processing each compilation unit in the normal ways, 
	 * we also need to do weaving for inter-type declarations.  This
	 * must be done before we use the signatures of these types to do any
	 * name binding.
	 */
	public void process(CompilationUnitDeclaration unit, int i) {
		EclipseFactory factory = 
			EclipseFactory.fromLookupEnvironment(lookupEnvironment);
		factory.showMessage(IMessage.INFO, "compiling " + new String(unit.getFileName()), null, null);
		super.process(unit, i);
				
		factory.finishedCompilationUnit(unit);
	}

	public BcelWorld getWorld() {
		return world;
	}
	
	public BcelWeaver getWeaver() {
		return weaver;
	}
	
	private void init(INameEnvironment nameEnvironment, IProblemFactory problemFactory, 
					  IMessageHandler handler, boolean inEclipse) {
		AjProblemReporter pr =
			new AjProblemReporter(DefaultErrorHandlingPolicies.proceedWithAllProblems(),
					this.options, problemFactory);
		
		problemReporter = pr;
		
		AjLookupEnvironment le =
			new AjLookupEnvironment(this, options, pr, nameEnvironment);
		
		lookupEnvironment = le;
		
		world = initBcelWorld(nameEnvironment,handler);
		weaver = new BcelWeaver(world);
		
		factory = inEclipse ? new InsideEclipseEclipseFactory(le,world) : new EclipseFactory(le);
		le.factory = factory;
		pr.factory = factory;
	}
	

	/**
	 * This method doesn't quite belong here, need to move to new home.... it's used by e.g. CompletionEngine 
	 * and SelectionEngine that work with a Parser and there own lookup environment, but don't create a full-fledged
	 * compiler.
	 * @param ajenv
	 * @param nEnv
	 * @param handler
	 */
	public static void initAjLookupEnvironment(AjLookupEnvironment ajenv, INameEnvironment nEnv, IMessageHandler handler) {
		BcelWorld world = initBcelWorld(nEnv,handler);
		ajenv.factory =  new InsideEclipseEclipseFactory(ajenv,world);		 
	}
	
	private static BcelWorld initBcelWorld(INameEnvironment nameEnvironment, IMessageHandler handler) {
		BcelWorld bcelWorld = new BcelWorld(getClasspathForWeaver(nameEnvironment), handler);
		// TODO AspectJ options setting...
		bcelWorld.setXnoInline(false);
		bcelWorld.getLint().loadDefaultProperties();

		//check for org.aspectj.runtime.JoinPoint
		bcelWorld.resolve("org.aspectj.lang.JoinPoint");
		return bcelWorld;
	}
	

	private static List getClasspathForWeaver(INameEnvironment nameEnvironment) {
		List cp = new ArrayList();
		String[] cpes = nameEnvironment.getExternalClasspath();
		for (int i = 0; i < cpes.length; i++) {
			cp.add(cpes[i]);
		}
		return cp;
	}
		
}
